@@ -1,13 +1,11 @@
 /*
   +----------------------------------------------------------------------+
-  | PHP Version 7                                                        |
-  +----------------------------------------------------------------------+
-  | Copyright (c) 1997-2018 The PHP Group                                |
+  | Copyright (c) The PHP Group                                          |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
-  | http://www.php.net/license/3_01.txt                                  |
+  | https://www.php.net/license/3_01.txt                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
@@ -22,15 +20,13 @@
 
 #include "php.h"
 
-#if DBA_QDBM
+#ifdef DBA_QDBM
 #include "php_qdbm.h"
 
 #ifdef QDBM_INCLUDE_FILE
 #include QDBM_INCLUDE_FILE
 #endif
 
-#define QDBM_DATA dba_qdbm_data *dba = info->dbf
-
 typedef struct {
 	DEPOT *dbf;
 } dba_qdbm_data;
@@ -41,16 +37,16 @@ DBA_OPEN_FUNC(qdbm)
 
 	switch(info->mode) {
 		case DBA_READER:
-			dbf = dpopen(info->path, DP_OREADER, 0);
+			dbf = dpopen(ZSTR_VAL(info->path), DP_OREADER, 0);
 			break;
 		case DBA_WRITER:
-			dbf = dpopen(info->path, DP_OWRITER, 0);
+			dbf = dpopen(ZSTR_VAL(info->path), DP_OWRITER, 0);
 			break;
 		case DBA_CREAT:
-			dbf = dpopen(info->path, DP_OWRITER | DP_OCREAT, 0);
+			dbf = dpopen(ZSTR_VAL(info->path), DP_OWRITER | DP_OCREAT, 0);
 			break;
 		case DBA_TRUNC:
-			dbf = dpopen(info->path, DP_OWRITER | DP_OCREAT | DP_OTRUNC, 0);
+			dbf = dpopen(ZSTR_VAL(info->path), DP_OWRITER | DP_OCREAT | DP_OTRUNC, 0);
 			break;
 		default:
 			return FAILURE;
@@ -69,7 +65,7 @@ DBA_OPEN_FUNC(qdbm)
 
 DBA_CLOSE_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 
 	dpclose(dba->dbf);
 	pefree(dba, info->flags & DBA_PERSISTENT);
@@ -77,30 +73,30 @@ DBA_CLOSE_FUNC(qdbm)
 
 DBA_FETCH_FUNC(qdbm)
 {
-	QDBM_DATA;
-	char *value, *new = NULL;
+	dba_qdbm_data *dba = info->dbf;
+	char *value;
 	int value_size;
+	zend_string *fetched_val = NULL;
 
-	value = dpget(dba->dbf, key, keylen, 0, -1, &value_size);
+	value = dpget(dba->dbf, ZSTR_VAL(key), ZSTR_LEN(key), 0, -1, &value_size);
 	if (value) {
-		if (newlen) *newlen = value_size;
-		new = estrndup(value, value_size);
+		fetched_val = zend_string_init(value, value_size, /* persistent */ false);
 		free(value);
 	}
 
-	return new;
+	return fetched_val;
 }
 
 DBA_UPDATE_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 
-	if (dpput(dba->dbf, key, keylen, val, vallen, mode == 1 ? DP_DKEEP : DP_DOVER)) {
+	if (dpput(dba->dbf, ZSTR_VAL(key), ZSTR_LEN(key), ZSTR_VAL(val), ZSTR_LEN(val), mode == 1 ? DP_DKEEP : DP_DOVER)) {
 		return SUCCESS;
 	}
 
 	if (dpecode != DP_EKEEP) {
-		php_error_docref2(NULL, key, val, E_WARNING, "%s", dperrmsg(dpecode));
+		php_error_docref(NULL, E_WARNING, "%s", dperrmsg(dpecode));
 	}
 
 	return FAILURE;
@@ -108,10 +104,10 @@ DBA_UPDATE_FUNC(qdbm)
 
 DBA_EXISTS_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 	char *value;
 
-	value = dpget(dba->dbf, key, keylen, 0, -1, NULL);
+	value = dpget(dba->dbf, ZSTR_VAL(key), ZSTR_LEN(key), 0, -1, NULL);
 	if (value) {
 		free(value);
 		return SUCCESS;
@@ -122,48 +118,48 @@ DBA_EXISTS_FUNC(qdbm)
 
 DBA_DELETE_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 
-	return dpout(dba->dbf, key, keylen) ? SUCCESS : FAILURE;
+	return dpout(dba->dbf, ZSTR_VAL(key), ZSTR_LEN(key)) ? SUCCESS : FAILURE;
 }
 
 DBA_FIRSTKEY_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 	int value_size;
-	char *value, *new = NULL;
+	char *value;
+	zend_string *key = NULL;
 
 	dpiterinit(dba->dbf);
 
 	value = dpiternext(dba->dbf, &value_size);
 	if (value) {
-		if (newlen) *newlen = value_size;
-		new = estrndup(value, value_size);
+		key = zend_string_init(value, value_size, /* persistent */ false);
 		free(value);
 	}
 
-	return new;
+	return key;
 }
 
 DBA_NEXTKEY_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 	int value_size;
-	char *value, *new = NULL;
+	char *value;
+	zend_string *key = NULL;
 
 	value = dpiternext(dba->dbf, &value_size);
 	if (value) {
-		if (newlen) *newlen = value_size;
-		new = estrndup(value, value_size);
+		key = zend_string_init(value, value_size, /* persistent */ false);
 		free(value);
 	}
 
-	return new;
+	return key;
 }
 
 DBA_OPTIMIZE_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 
 	dpoptimize(dba->dbf, 0);
 	return SUCCESS;
@@ -171,7 +167,7 @@ DBA_OPTIMIZE_FUNC(qdbm)
 
 DBA_SYNC_FUNC(qdbm)
 {
-	QDBM_DATA;
+	dba_qdbm_data *dba = info->dbf;
 
 	dpsync(dba->dbf);
 	return SUCCESS;
@@ -183,12 +179,3 @@ DBA_INFO_FUNC(qdbm)
 }
 
 #endif
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: sw=4 ts=4 fdm=marker
- * vim<600: sw=4 ts=4
- */
