@@ -5,7 +5,7 @@
    | This source file is subject to version 3.01 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
    | available through the world-wide-web at the following url:           |
-   | http://www.php.net/license/3_01.txt                                  |
+   | https://www.php.net/license/3_01.txt                                 |
    | If you did not receive a copy of the PHP license and are unable to   |
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
@@ -26,7 +26,7 @@
 #include <errno.h>
 
 static int fpm_event_port_init(int max);
-static int fpm_event_port_clean();
+static int fpm_event_port_clean(void);
 static int fpm_event_port_wait(struct fpm_event_queue_s *queue, unsigned long int timeout);
 static int fpm_event_port_add(struct fpm_event_s *ev);
 static int fpm_event_port_remove(struct fpm_event_s *ev);
@@ -47,7 +47,7 @@ static int pfd = -1;
 
 #endif /* HAVE_PORT */
 
-struct fpm_event_module_s *fpm_event_port_module() /* {{{ */
+struct fpm_event_module_s *fpm_event_port_module(void) /* {{{ */
 {
 #ifdef HAVE_PORT
 	return &port_module;
@@ -90,7 +90,7 @@ static int fpm_event_port_init(int max) /* {{{ */
 /*
  * Clean the module
  */
-static int fpm_event_port_clean() /* {{{ */
+static int fpm_event_port_clean(void)
 {
 	if (pfd > -1) {
 		close(pfd);
@@ -105,7 +105,6 @@ static int fpm_event_port_clean() /* {{{ */
 	nevents = 0;
 	return 0;
 }
-/* }}} */
 
 /*
  * wait for events or timeout
@@ -120,7 +119,7 @@ static int fpm_event_port_wait(struct fpm_event_queue_s *queue, unsigned long in
 	t.tv_sec = (int)(timeout / 1000);
 	t.tv_nsec = (timeout % 1000) * 1000 * 1000;
 
-	/* wait for inconming event or timeout. We want at least one event or timeout */
+	/* wait for incoming event or timeout. We want at least one event or timeout */
 	nget = 1;
 	events[0].portev_user = (void *)-1; /* so we can double check that an event was returned */
 
@@ -145,14 +144,20 @@ static int fpm_event_port_wait(struct fpm_event_queue_s *queue, unsigned long in
 	}
 
 	for (i = 0; i < nget; i++) {
+		struct fpm_event_s *ev;
 
 		/* do we have a ptr to the event ? */
 		if (!events[i].portev_user) {
 			continue;
 		}
 
+		ev = (struct fpm_event_s *)events[i].portev_user;
+
+		/* re-associate for next event */
+		fpm_event_port_add(ev);
+
 		/* fire the event */
-		fpm_event_fire((struct fpm_event_s *)events[i].portev_user);
+		fpm_event_fire(ev);
 
 		/* sanity check */
 		if (fpm_globals.parent_pid != getpid()) {
